import asyncio
import datetime
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from functools import singledispatchmethod
from importlib import metadata

import httpx

from wedne.commands import CommandSchema


@dataclass
class Watcher:
    endpoint: str
    delay: datetime.timedelta
    social_media_id: int
    command_processor: Callable[[CommandSchema | None], Awaitable[None]]

    async def __call__(self) -> None:
        version = metadata.version("wedne")
        client = httpx.AsyncClient(headers={"user-agent": f"wedne.client/{version}"})
        try:
            await self._watch(client)
        finally:
            await client.aclose()

    async def _watch(self, client: httpx.AsyncClient) -> None:
        data = {
            "social_media_id": self.social_media_id,
        }
        while True:
            try:
                response = await client.post(
                    self.endpoint,
                    json=data,
                )
                await self.process(response)
            except httpx.ConnectError as exc:
                await self.process(exc)
            await asyncio.sleep(self.delay.seconds)

    @singledispatchmethod
    async def process(self, response: httpx.Response) -> None:
        print(response)
        print(response.content)

        if response.is_server_error:
            print("Server error")
            return

        if response.is_client_error:
            print("Client error, update me")
            return

        raw_command = response.json()
        command = CommandSchema.parse_obj(raw_command) if raw_command else None
        await self.command_processor(command)

    @process.register
    async def _(self, exception: httpx.ConnectError) -> None:
        print(str(exception))



import asyncio
import datetime

import pytz
import telethon

from wedne.client.consts import WRITE_FIRST_DELAY
from wedne.client.shared_commands import SharedCommand
from wedne.commands import CommandSchema


def get_handler(shared_command: SharedCommand):
    async def handler(event: telethon.events.NewMessage.Event) -> None:
        print(event)
        print(f"got new {type(event)}, btw, my task is", await shared_command.read())
        command = await shared_command.read()
        if command is None:
            # no pending command
            print("no pending")
            return
        if command.when > datetime.datetime.now(pytz.utc):
            # it's too early by time
            print("too early by time")
            return
        if command.after is not None and command.after != event.from_id.user_id:
            # it's still early by letter order
            print("too early by letter order")
            return
        await event.reply(command.letter)
        await shared_command.clear()

    return handler


class ChatMonitor:
    def __init__(self, client: telethon.TelegramClient, shared_command: SharedCommand):
        self.client = client
        self.shared_command = shared_command

    async def __call__(self, chat_id: int) -> None:
        self.client.add_event_handler(
            callback=get_handler(self.shared_command),
            event=telethon.events.NewMessage(
                chats=[chat_id],
                incoming=True,
                forwards=False,
            ),
        )


class TelegramTowerBuilder:
    def __init__(self, session: str, api_id: int, api_hash: str, chat_id: int):
        self.client = telethon.TelegramClient(
            session,
            api_id,
            api_hash,
        )
        self.chat_id = chat_id
        self.shared_command = SharedCommand()

    async def start(self) -> None:
        await self.client.start()  # type: ignore

    async def who_am_i(self) -> int:
        return (await self.client.get_me()).id  # type: ignore

    async def monitor(self) -> None:
        await ChatMonitor(self.client, self.shared_command)(
            self.chat_id,
        )

    async def process_command(self, command: CommandSchema | None):
        if command is None:
            await self.shared_command.clear()
            return
        await asyncio.sleep(
            (command.when - datetime.datetime.now(pytz.utc)).seconds,
        )
        if command.after is None:
            # if it's the first letter, wait here and write a message
            await asyncio.sleep(WRITE_FIRST_DELAY.seconds)
            await self.client.send_message(self.chat_id, command.letter)
        else:
            # or wait for a necessary message in ChatMonitor
            await self.shared_command.write(command)



from pathlib import Path

import pydantic


class Settings(pydantic.BaseSettings):
    minutes_to_order: int
    seconds_of_delay: int
    minutes_of_last_activity: int
    app_database: Path
    job_queue_database: Path
    tower: str = "ITSWEDNESDAYMYDUDES"
    host: str
    port: int
    reload: bool = False

    class Config:
        env_file = ".env"
        env_prefix = "wedne_"


settings = Settings()  # type: ignore



from datetime import datetime, timedelta

import fastapi
import pydantic
import pytz

from wedne.commands import CommandSchema, PreciseCommandSchema
from wedne.server.db import VisitDAO
from wedne.server.settings import settings
from wedne.server.tasks import create_new_visit

router = fastapi.APIRouter()


class VisitInputSchema(pydantic.BaseModel):
    social_media_id: int


@router.post("/")
def take_a_visit(
    data: VisitInputSchema,
    dao: VisitDAO = fastapi.Depends(),
) -> PreciseCommandSchema | None:
    now = datetime.now(pytz.utc)
    create_new_visit(
        time=now,
        social_media_id=data.social_media_id,
    )
    raw_command = dao.get_command(
        from_=now - timedelta(seconds=settings.minutes_to_order),
        social_media_id=data.social_media_id,
    )
    if raw_command is None:
        return None
    command = CommandSchema.parse_obj(raw_command)
    return PreciseCommandSchema(
        expected_in=command.when - now,
        **command.dict(),
    )



from collections.abc import Callable, Hashable, Iterable, Sequence
from typing import TypeVar

T_ = TypeVar("T_")


def distinct_on(seq: Iterable[T_], key: Callable[[T_], Hashable]) -> Sequence[T_]:
    unique = {key(element): element for element in seq}
    return list(unique.values())



import datetime
from typing import Annotated

import pydantic


class CommandSchema(pydantic.BaseModel):
    letter: Annotated[str, pydantic.constr(min_length=1, max_length=1)]
    after: int | None = pydantic.Field(
        description="social_media_id user before current command",
    )
    when: datetime.datetime


class PreciseCommandSchema(CommandSchema):
    expected_in: datetime.timedelta



import random
from datetime import datetime, timedelta

import huey
import pytz

from wedne.commands import CommandSchema
from wedne.server.db import VisitDAO, VisitModel
from wedne.server.settings import settings
from wedne.utils import distinct_on

job_queue = huey.SqliteHuey(str(settings.job_queue_database))


@job_queue.task(priority=10)
def create_new_visit(time: datetime, social_media_id: int) -> None:
    dao = VisitDAO()
    dao.create(time, social_media_id)


def choose_randoms_to_order(dao: VisitDAO) -> list[VisitModel]:
    """Raises ValueError if there are not sufficient candidates."""
    active = dao.get_from(
        datetime.now(pytz.utc) - timedelta(minutes=settings.minutes_of_last_activity),
    )
    return random.sample(
        distinct_on(active, lambda vis: vis.social_media_id),
        k=len(settings.tower),
    )


@job_queue.periodic_task(
    huey.crontab(minute=f"*/{settings.minutes_to_order}", strict=True),
    priority=20,
)
def order_new_commands() -> None:
    dao = VisitDAO()
    previous_social_media_id = None
    try:
        chosen = choose_randoms_to_order(dao)
    except ValueError:
        raise
    when = datetime.now(pytz.utc) + timedelta(seconds=settings.seconds_of_delay)
    for idx, letter in enumerate(settings.tower):
        command = CommandSchema(
            letter=letter,
            after=previous_social_media_id,
            when=when,
        )
        dao.update_command(chosen[idx].id, command.dict())
        previous_social_media_id = chosen[idx].social_media_id
